// Michael J. Wilson
// Music 256a Fall 2010
// Homework #2

#include "RtAudio.h"
#include "RtMidi.h"
#include "UnitDelay.h"
#include "UnitDelayFeedback.h"
#include "UnitGen.h"
#include "UnitKarplusStrong.h"
#include "MidiKarplusStrongInstrument.h"
#include "Utility.h"
#include <iostream>

/// For debugging
const bool DEBUG = false;

//////////////////////////////////////////////////////////////////////
/// Audio callback
//////////////////////////////////////////////////////////////////////
int audio_callback(void *outputBuffer, void *inputBuffer, unsigned int nBufferFrames, double streamTime, RtAudioStreamStatus status, void *userData)
{
    unsigned int i;
    unsigned int j;
    double* obuffer = (double*) outputBuffer;
    double* ibuffer = inputBuffer != NULL ? (double*)inputBuffer : NULL;
    UnitGen* unit_gen = (UnitGen*) userData;
    double output_sample;
    double input_sample;

    if (status)
    {
        std::cout << "Stream over/underflow detected!" << std::endl;
    }

    if (inputBuffer != NULL)
    {
        ibuffer = (double*) inputBuffer;
        for (i=0; i<nBufferFrames; i++)
        {
            input_sample = (*ibuffer++);
            output_sample = unit_gen->get_next_sample(&input_sample);

            for (j=0; j<Utility::OUTPUT_CHANNELS; j++)
            {
                *obuffer++ = output_sample;
            }
        }
    }
    else
    {
        for (i=0; i<nBufferFrames; i++)
        {
            output_sample = unit_gen->get_next_sample();
            for (j=0; j<Utility::OUTPUT_CHANNELS; j++)
            {
                *obuffer++ = output_sample;
            }
        }
    }

    return(0);
}

//////////////////////////////////////////////////////////////////////
/// Midi callback
//////////////////////////////////////////////////////////////////////
void midi_callback( double deltatime, std::vector< unsigned char > *message, void *userData )
{
    unsigned int nBytes = message->size();
    MidiKarplusStrongInstrument* instrument = (MidiKarplusStrongInstrument*) userData;
    unsigned int control_message;
    unsigned int velocity;
    unsigned int note;

    if (DEBUG)
    {
        for ( unsigned int i=0; i<nBytes; i++ )
        {
            std::cout << "Byte " << i << " = " << (int)message->at(i) << ", ";
        }
        if ( nBytes > 0 )
        {
            std::cout << "stamp = " << deltatime << std::endl;
        }
    }

    // We need at least a note on/of and a note name.  So we are only
    // interested in messages with at least two bytes.
    if (message->size() > 1)
    {
        // TODO this is probably not 64-bit compliant...
        control_message = 0xF0 & (unsigned int)message->at(0);
        note = (unsigned int)message->at(1);
        if (message->size() > 2)
        {
            velocity = (unsigned int)message->at(2);
        }
        else
        {
            // If we get a nonstandard message without velocity
            // information, just play at full velocity.
            velocity = 127;
        }

        if (control_message == 0x90)
        {
            if (velocity > 0)
            {
                // Note on
                instrument->play(note, velocity);
            }
            else
            {
                // If velocity is 0, turn it off
                instrument->release(note);
            }
        }
        else if (control_message == 0x80)
        {
            // Note off
            instrument->release(note);
        }
    }
}

//////////////////////////////////////////////////////////////////////
/// Handle main input
//////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
    bool error_occured = false;

    //////////////////////////////////////////////////
    // Parameters
    unsigned int lowest_note = 0;
    unsigned int midi_port = 0;
    unsigned int polyphony = 32;

    //////////////////////////////////////////////////
    // Parse command-line options
    const unsigned int option_count = 2;
    unsigned int* option_values[] =
        {
            &midi_port,
            &polyphony
        };
    const std::string option_strings[] =
        {
            "--port",
            "--poly"
        };
    for (int i = 1; i < argc; ++i)
    {
        if ((i+1) < argc)
        {
            for (unsigned int j = 0; j < option_count; ++j)
            {
                if (option_strings[j].compare(argv[i]) == 0)
                {
                    if (!(std::istringstream(argv[i+1]) >> *option_values[j]))
                    {
                        error_occured = true;
                    }
                    else
                    {
                        i++;
                    }
                }
            }
        }
        else
        {
            error_occured = true;
        }
    }
    if (error_occured)
    {
        // Print usage message
        std::cout << Utility::PROGRAM_NAME << " [OPTIONS] " << std::endl
                  << "    OPTIONS:" << std::endl
                  << "        --port <value> Midi input port to use" << std::endl
                  << "        --poly <value> Amount of polyphony to use" << std::endl
                  << "    EXAMPLES:" << std::endl
                  << "        " << Utility::PROGRAM_NAME << " --port 3 --poly 32" << std::endl
                  << "        " << Utility::PROGRAM_NAME << " --poly 12 --port 0" << std::endl;
    }

    //////////////////////////////////////////////////
    // Set up instrument
    MidiKarplusStrongInstrument* instrument = 0;
    if (!error_occured)
    {
        instrument = new MidiKarplusStrongInstrument(polyphony, lowest_note, 0.99999);
    }

    //////////////////////////////////////////////////
    // Set up audio
    RtAudio dac;
    if (dac.getDeviceCount() < 1)
    {
        std::cout << std::endl << "No audio devices found!" << std::endl;
        error_occured = true;
    }
    if (!error_occured)
    {
        unsigned int bufferFrames = 128;

        RtAudio::StreamParameters outputParameters;
        outputParameters.deviceId = dac.getDefaultOutputDevice();
        outputParameters.nChannels = Utility::OUTPUT_CHANNELS;
        outputParameters.firstChannel = 0;

        std::cout << "Sample rate: " << Utility::SAMPLE_RATE << std::endl
                  << "Frames: " << bufferFrames << std::endl
                  << "Output channels: " << Utility::OUTPUT_CHANNELS << std::endl;
        try
        {
            dac.openStream(&outputParameters,
                           NULL,
                           RTAUDIO_FLOAT64,
                           Utility::SAMPLE_RATE,
                           &bufferFrames,
                           &audio_callback,
                           instrument);
            dac.startStream();
        }
        catch (RtError& e)
        {
            e.printMessage();
            error_occured = true;
        }
    }

    //////////////////////////////////////////////////
    // Set up MIDI
    RtMidiIn* midiin = 0;
    if (!error_occured)
    {
        try
        {
            midiin = new RtMidiIn();
        }
        catch ( RtError &error )
        {
            error.printMessage();
            error_occured = true;
        }

        if (!error_occured)
        {
            // Open the midi port
            std::cout << "Midi port count: " << midiin->getPortCount() << std::endl
                      <<"Opening port: " << midi_port << std::endl;
            try
            {
                midiin->openPort(midi_port);
            }
            catch ( RtError &error )
            {
                error.printMessage();
                error_occured = true;
            }
        }

        if (!error_occured)
        {
            // Set our callback function.  This should be done immediately after
            // opening the port to avoid having incoming messages written to the
            // queue instead of sent to the callback function.
            midiin->setCallback(&midi_callback,
                                instrument);

            // Don't ignore sysex, timing, or active sensing messages.
            midiin->ignoreTypes( false, false, false );
        }
    }

    //////////////////////////////////////////////////
    // Play and listen for input
    if (!error_occured)
    {
        char input;
        std::cout << std::endl
                  << "Send Midi messages to hear output." << std::endl
                  << "Press [Enter] to quit." << std::endl;
        std::cin.get(input);
        try
        {
            // Stop the stream
            dac.stopStream();
        }
        catch (RtError& e)
        {
            e.printMessage();
            error_occured = true;
        }
    }

    //////////////////////////////////////////////////
    // Cleanup
    if (dac.isStreamOpen())
    {
        dac.closeStream();
    }
    delete midiin;
    delete instrument;

    return (error_occured ? 1 : 0);
}
